@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* history.c: command history management */
-/* (C) 2007-2013 magicant */
+/* (C) 2007-2015 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -213,7 +213,7 @@ histentry_T *new_entry(unsigned number, time_t time, const char *line)
 	remove_entry(ashistentry(histlist.Oldest));
 
     histentry_T *new = xmallocs(sizeof *new,
-	    strlen(line) + 1, sizeof *new->value);
+	    add(strlen(line), 1), sizeof *new->value);
     new->Prev = histlist.Newest;
     new->Next = Histlist;
     histlist.Newest = new->Prev->next = &new->link;
@@ -422,8 +422,8 @@ void add_histfile_pid(pid_t pid)
 	if (histfilepids.pids[i] == pid)
 	    return;  /* don't add if already added */
 
-    histfilepids.pids = xreallocn(histfilepids.pids,
-	    histfilepids.count + 1, sizeof *histfilepids.pids);
+    histfilepids.pids = xrealloce(histfilepids.pids,
+	    histfilepids.count, 1, sizeof *histfilepids.pids);
     histfilepids.pids[histfilepids.count++] = pid;
 }
 
@@ -627,7 +627,7 @@ bool try_read_line(FILE *restrict f, xwcsbuf_T *restrict buf)
 	}
 	if (buf->length > LINE_MAX)
 	    return false;  /* Too long line. Give up. */
-	wb_ensuremax(buf, buf->length + 80);
+	wb_ensuremax(buf, add(buf->length, 80));
     }
     return false;
 }
@@ -1524,7 +1524,7 @@ int fc_edit_and_exec_entries(
     pid_t cpid;
     int savelaststatus;
 
-    fd = create_temporary_file(&temp, S_IRUSR | S_IWUSR);
+    fd = create_temporary_file(&temp, ".sh", S_IRUSR | S_IWUSR);
     if (fd < 0) {
 	xerror(errno, Ngt("cannot create a temporary file to edit history"));
 	goto error1;
