@@ -479,7 +479,7 @@ int wait_for_job(size_t jobnumber, bool return_on_stop,
  * The exit status is assigned to `laststatus' in any case. */
 wchar_t **wait_for_child(pid_t cpid, pid_t cpgid, bool return_on_stop)
 {
-    job_T *job = xmalloc(sizeof *job + sizeof *job->j_procs);
+    job_T *job = xmalloc(add(sizeof *job, sizeof *job->j_procs));
     job->j_pgid = cpgid;
     job->j_status = JS_RUNNING;
     job->j_statuschanged = false;
@@ -1155,9 +1155,9 @@ int continue_job(size_t jobnumber, job_T *job, bool fg)
 
     wchar_t *name = get_job_name(job);
     if (fg && posixly_correct)
-	printf("%ls\n", name);
+	xprintf("%ls\n", name);
     else
-	printf("[%zu] %ls\n", jobnumber, name);
+	xprintf("[%zu] %ls\n", jobnumber, name);
     if (name != job->j_procs[0].pr_name)
 	free(name);
 
