@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* parser.c: syntax parser */
-/* (C) 2007-2012 magicant */
+/* (C) 2007-2015 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -32,9 +32,7 @@
 #include <unistd.h>
 #include <wchar.h>
 #include <wctype.h>
-#if YASH_ENABLE_ALIAS
-# include "alias.h"
-#endif
+#include "alias.h"
 #include "expand.h"
 #include "input.h"
 #include "option.h"
@@ -81,8 +79,7 @@ void pipesfree(pipeline_T *p)
 void comsfree(command_T *c)
 {
     while (c != NULL) {
-	c->refcount--;
-	if (c->refcount > 0)
+	if (!refcount_decrement(&c->refcount))
 	    break;
 
 	redirsfree(c->c_redirs);
@@ -245,7 +242,11 @@ void embedcmdfree(embedcmd_T c)
 
 /********** Auxiliary Functions for Parser **********/
 
-static wchar_t *skip_name(const wchar_t *s)
+static wchar_t *skip_name(const wchar_t *s, bool predicate(wchar_t))
+    __attribute__((pure,nonnull));
+static bool is_name_by_predicate(const wchar_t *s, bool predicate(wchar_t))
+    __attribute__((pure,nonnull));
+static bool is_portable_name(const wchar_t *s)
     __attribute__((pure,nonnull));
 
 
@@ -281,19 +282,33 @@ bool is_name_char(wchar_t c)
 
 /* Skips an identifier at the head of the specified string and returns a
  * pointer to the character right after the identifier in the string.
- * If there is no identifier, the argument `s' is simply returned. */
-wchar_t *skip_name(const wchar_t *s)
+ * If there is no identifier, the argument `s' is simply returned. `predicate`
+ * determines if a character is valid. */
+wchar_t *skip_name(const wchar_t *s, bool predicate(wchar_t))
 {
     if (!iswdigit(*s))
-	while (is_name_char(*s))
+	while (predicate(*s))
 	    s++;
     return (wchar_t *) s;
 }
 
+/* Returns true iff the specified string constitutes a valid identifier that
+ * is made up of characters accepted by `predicate'. */
+bool is_name_by_predicate(const wchar_t *s, bool predicate(wchar_t))
+{
+    return s[0] != L'\0' && skip_name(s, predicate)[0] == L'\0';
+}
+
+/* Returns true iff the specified string constitutes a valid portable name. */
+bool is_portable_name(const wchar_t *s)
+{
+    return is_name_by_predicate(s, is_portable_name_char);
+}
+
 /* Returns true iff the specified string constitutes a valid identifier. */
 bool is_name(const wchar_t *s)
 {
-    return s[0] != L'\0' && skip_name(s)[0] == L'\0';
+    return is_name_by_predicate(s, is_name_char);
 }
 
 /* Returns true iff the string is a reserved word. */
@@ -351,10 +366,8 @@ typedef struct parsestate_T {
     struct xwcsbuf_T src;
     size_t index;
     struct plist_T pending_heredocs;
-#if YASH_ENABLE_ALIAS
     bool enable_alias, reparse;
     struct aliaslist_T *aliases;
-#endif
 } parsestate_T;
 /* info: contains parameter that affect the behavior of parsing.
  * error: set to true when an parsing error occurs.
@@ -443,10 +456,8 @@ static wordunit_T *parse_cmdsubst_in_paren(parsestate_T *ps)
     __attribute__((nonnull,malloc,warn_unused_result));
 static embedcmd_T extract_command_in_paren(parsestate_T *ps)
     __attribute__((nonnull,warn_unused_result));
-#if YASH_ENABLE_ALIAS
 static wchar_t *extract_command_in_paren_unparsed(parsestate_T *ps)
     __attribute__((nonnull,malloc,warn_unused_result));
-#endif
 static wordunit_T *parse_cmdsubst_in_backquote(parsestate_T *ps)
     __attribute__((nonnull,malloc,warn_unused_result));
 static wordunit_T *tryparse_arith(parsestate_T *ps)
@@ -524,11 +535,9 @@ parseresult_T read_and_parse(parseparam_T *info, and_or_T **restrict resultp)
 	.info = info,
 	.error = false,
 	.index = 0,
-#if YASH_ENABLE_ALIAS
 	.enable_alias = info->enable_alias,
 	.reparse = false,
 	.aliases = NULL,
-#endif
     };
 
     if (ps.info->interactive) {
@@ -544,9 +553,7 @@ parseresult_T read_and_parse(parseparam_T *info, and_or_T **restrict resultp)
 
     wb_destroy(&ps.src);
     pl_destroy(&ps.pending_heredocs);
-#if YASH_ENABLE_ALIAS
     destroy_aliaslist(ps.aliases);
-#endif
 
     switch (ps.info->lastinputresult) {
 	case INPUT_OK:
@@ -883,12 +890,10 @@ and_or_T *parse_command_list(parsestate_T *ps, bool toeol)
 	    *lastp = ao;
 	    lastp = &ao->next;
 	}
-#if YASH_ENABLE_ALIAS
 	if (ps->reparse) {
 	    assert(ao == NULL);
 	    continue;
 	}
-#endif
 
 	need_separator = true;
 	ensure_buffer(ps, 2);
@@ -901,9 +906,7 @@ and_or_T *parse_command_list(parsestate_T *ps, bool toeol)
     }
     if (!toeol)
 	ps->error |= saveerror;
-#if YASH_ENABLE_ALIAS
     ps->reparse = false;
-#endif
     return first;
 }
 
@@ -922,12 +925,10 @@ and_or_T *parse_compound_list(parsestate_T *ps)
 and_or_T *parse_and_or_list(parsestate_T *ps)
 {
     pipeline_T *p = parse_pipelines_in_and_or(ps);
-#if YASH_ENABLE_ALIAS
     if (ps->reparse) {
 	assert(p == NULL);
 	return NULL;
     }
-#endif
 
     and_or_T *result = xmalloc(sizeof *result);
     result->next = NULL;
@@ -951,7 +952,6 @@ pipeline_T *parse_pipelines_in_and_or(parsestate_T *ps)
 	    *lastp = p;
 	    lastp = &p->next;
 	}
-#if YASH_ENABLE_ALIAS
 	if (ps->reparse) {
 	    assert(p == NULL);
 	    if (first != NULL)
@@ -959,7 +959,6 @@ pipeline_T *parse_pipelines_in_and_or(parsestate_T *ps)
 	    else
 		break;
 	}
-#endif
 
 	ensure_buffer(ps, 2);
 	if (ps->src.contents[ps->index] == L'&'
@@ -972,9 +971,7 @@ pipeline_T *parse_pipelines_in_and_or(parsestate_T *ps)
 	    break;
 	}
 	ps->index += 2;
-#if YASH_ENABLE_ALIAS
 next:
-#endif
 	skip_to_next_token(ps);
     }
     return first;
@@ -992,25 +989,19 @@ pipeline_T *parse_pipeline(parsestate_T *ps)
     if (has_token(ps, L"!")) {
 	neg = true;
 	ps->index += 1;
-#if YASH_ENABLE_ALIAS
 	do {
-#endif
 	    skip_blanks_and_comment(ps);
 	    c = parse_commands_in_pipeline(ps);
-#if YASH_ENABLE_ALIAS
 	    if (ps->reparse)
 		assert(c == NULL);
 	} while (ps->reparse);
-#endif
     } else {
 	neg = false;
 	c = parse_commands_in_pipeline(ps);
-#if YASH_ENABLE_ALIAS
 	if (ps->reparse) {
 	    assert(c == NULL);
 	    return NULL;
 	}
-#endif
     }
 
     pipeline_T *result = xmalloc(sizeof *result);
@@ -1034,7 +1025,6 @@ command_T *parse_commands_in_pipeline(parsestate_T *ps)
 	    *lastp = c;
 	    lastp = &c->next;
 	}
-#if YASH_ENABLE_ALIAS
 	if (ps->reparse) {
 	    assert(c == NULL);
 	    if (first != NULL)
@@ -1042,7 +1032,6 @@ command_T *parse_commands_in_pipeline(parsestate_T *ps)
 	    else
 		break;
 	}
-#endif
 
 	ensure_buffer(ps, 2);
 	if (ps->src.contents[ps->index] == L'|' &&
@@ -1051,9 +1040,7 @@ command_T *parse_commands_in_pipeline(parsestate_T *ps)
 	} else {
 	    break;
 	}
-#if YASH_ENABLE_ALIAS
 next:
-#endif
 	skip_to_next_token(ps);
     }
     return first;
@@ -1064,9 +1051,7 @@ next:
  * NULL is returned. */
 command_T *parse_command(parsestate_T *ps)
 {
-#if YASH_ENABLE_ALIAS
     ps->reparse = false;
-#endif
 
     /* Note: `check_closing_token' calls `ensure_buffer(ps, 5)'. */
     const wchar_t *t = check_closing_token(ps);
@@ -1095,7 +1080,6 @@ command_T *parse_command(parsestate_T *ps)
     if (t != NULL)
 	return parse_compound_command(ps, t);
 
-#if YASH_ENABLE_ALIAS
     if (ps->enable_alias && count_name_length(ps, is_alias_name_char) > 0) {
 	substaliasflags_T flags = AF_NONGLOBAL | AF_NORECUR;
 	if (substitute_alias(&ps->src, ps->index, &ps->aliases, flags)) {
@@ -1103,7 +1087,6 @@ command_T *parse_command(parsestate_T *ps)
 	    return NULL;
 	}
     }
-#endif
 
     command_T *result = tryparse_function(ps);
     if (result != NULL)
@@ -1152,13 +1135,11 @@ redir_T **parse_assignments_and_redirects(parsestate_T *ps, command_T *c)
 	} else {
 	    break;
 	}
-#if YASH_ENABLE_ALIAS
 	if (ps->enable_alias && count_name_length(ps, is_alias_name_char) > 0) {
 	    substitute_alias(
 		    &ps->src, ps->index, &ps->aliases, AF_NONGLOBAL);
 	    skip_blanks_and_comment(ps);
 	}
-#endif
     }
     return redirlastp;
 }
@@ -1181,16 +1162,12 @@ void **parse_words_and_redirects(
     pl_init(&wordlist);
     while (ensure_buffer(ps, 1),
 	    !is_command_delimiter_char(ps->src.contents[ps->index])) {
-#if YASH_ENABLE_ALIAS
 	if (!first && ps->enable_alias) {
 	    if (count_name_length(ps, is_alias_name_char) > 0) {
 		substitute_alias(&ps->src, ps->index, &ps->aliases, 0);
 		skip_blanks_and_comment(ps);
 	    }
 	}
-#else
-	(void) first;
-#endif
 	if ((redir = tryparse_redirect(ps)) != NULL) {
 	    *redirlastp = redir;
 	    redirlastp = &redir->next;
@@ -1211,11 +1188,9 @@ void **parse_words_and_redirects(
 void parse_redirect_list(parsestate_T *ps, redir_T **lastp)
 {
     for (;;) {
-#if YASH_ENABLE_ALIAS
 	if (!posixly_correct && ps->enable_alias)
 	    if (count_name_length(ps, is_alias_name_char) > 0)
 		substitute_alias(&ps->src, ps->index, &ps->aliases, 0);
-#endif
 
 	redir_T *redir = tryparse_redirect(ps);
 	if (redir == NULL)
@@ -1410,7 +1385,6 @@ parse_command:
  * `type' specifies the type of aliases to be expanded.  */
 wordunit_T *parse_word(parsestate_T *ps, aliastype_T type)
 {
-#if YASH_ENABLE_ALIAS
     if (ps->enable_alias) {
 	switch (type) {
 	case AT_NONE:
@@ -1426,9 +1400,6 @@ wordunit_T *parse_word(parsestate_T *ps, aliastype_T type)
 	    break;
 	}
     }
-#else /* !YASH_ENABLE_ALIAS */
-    (void) type;
-#endif
 
     return parse_word_to(ps, is_token_delimiter_char);
 }
@@ -1827,13 +1798,10 @@ embedcmd_T extract_command_in_paren(parsestate_T *ps)
     save_pending_heredocs = ps->pending_heredocs;
     pl_init(&ps->pending_heredocs);
 
-#if YASH_ENABLE_ALIAS
     if (posixly_correct && ps->info->enable_alias) {
 	result.is_preparsed = false;
 	result.value.unparsed = extract_command_in_paren_unparsed(ps);
-    } else
-#endif
-    {
+    } else {
 	ps->index++;
 	result.is_preparsed = true;
 	result.value.preparsed = parse_compound_list(ps);
@@ -1845,7 +1813,6 @@ embedcmd_T extract_command_in_paren(parsestate_T *ps)
     return result;
 }
 
-#if YASH_ENABLE_ALIAS
 /* Parses commands between '(' and ')'.
  * The current position must be at the opening parenthesis L'(' when this
  * function is called. The position is advanced to the closing parenthesis
@@ -1865,7 +1832,6 @@ wchar_t *extract_command_in_paren_unparsed(parsestate_T *ps)
     ps->enable_alias = save_enable_alias;
     return result;
 }
-#endif
 
 /* Parses a command substitution enclosed by backquotes.
  * When this function is called, the current position must be at the character
@@ -2161,7 +2127,7 @@ command_T *parse_for(parsestate_T *ps)
     result->c_redirs = NULL;
 
     wchar_t *name = parse_word_as_wcs(ps);
-    if (!is_name(name)) {
+    if (!(posixly_correct ? is_portable_name : is_name)(name)) {
 	if (name[0] == L'\0')
 	    serror(ps, Ngt("an identifier is required after `for'"));
 	else
@@ -2419,7 +2385,8 @@ command_T *tryparse_function(parsestate_T *ps)
     if (iswdigit(ps->src.contents[ps->index]))
 	goto fail;
 
-    size_t namelen = count_name_length(ps, is_name_char);
+    size_t namelen = count_name_length(ps,
+	    posixly_correct ? is_portable_name_char : is_name_char);
     ps->index += namelen;
     if (namelen == 0 || !is_token_delimiter_char(ps->src.contents[ps->index]))
 	goto fail;
@@ -2495,6 +2462,8 @@ void read_heredoc_contents_without_expansion(parsestate_T *ps, redir_T *r)
 	} else {
 	    /* encountered EOF before reading an end-of-contents marker! */
 	    linelen = ps->src.length - ps->index;
+	    if (posixly_correct)
+		serror(ps, Ngt("the here-document is not closed"));
 	}
 	wb_ncat_force(&buf, &ps->src.contents[ps->index], linelen);
 	ps->index += linelen;
@@ -2523,9 +2492,12 @@ void read_heredoc_contents_with_expansion(parsestate_T *ps, redir_T *r)
     bool skiptab = (r->rd_type == RT_HERERT);
 
     while (!is_end_of_heredoc_contents(ps, eoc, skiptab)) {
+	size_t oldindex = ps->index;
 	lastp = parse_string_without_quotes(ps, true, true, lastp);
-	if (ps->src.contents[ps->index - 1] != L'\n') {
+	if (ps->index == oldindex || ps->src.contents[ps->index - 1] != L'\n') {
 	    /* encountered EOF before reading an end-of-contents marker! */
+	    if (posixly_correct)
+		serror(ps, Ngt("the here-document is not closed"));
 	    break;
 	}
     }
@@ -2544,7 +2516,7 @@ bool is_end_of_heredoc_contents(
 
     if (ps->src.contents[ps->index] == L'\0')
 	if (read_more_input(ps) != INPUT_OK)
-	    return true;
+	    return !posixly_correct;
     if (skiptab)
 	while (ps->src.contents[ps->index] == L'\t')
 	    ps->index++;
@@ -2639,11 +2611,9 @@ bool parse_string(parseparam_T *info, wordunit_T **restrict resultp)
 	.info = info,
 	.error = false,
 	.index = 0,
-#if YASH_ENABLE_ALIAS
 	.enable_alias = false,
 	.reparse = false,
 	.aliases = NULL,
-#endif
     };
     wb_init(&ps.src);
 
@@ -2656,10 +2626,8 @@ bool parse_string(parseparam_T *info, wordunit_T **restrict resultp)
 
     wb_destroy(&ps.src);
     pl_destroy(&ps.pending_heredocs);
-#if YASH_ENABLE_ALIAS
     assert(ps.aliases == NULL);
     //destroy_aliaslist(ps.aliases);
-#endif
     if (ps.info->lastinputresult != INPUT_EOF || ps.error) {
 	wordfree(*resultp);
 	return false;
@@ -3202,8 +3170,11 @@ void print_word(struct print *restrict pr, const wordunit_T *restrict wu,
 		break;
 	    case WT_CMDSUB:
 		wb_cat(&pr->buffer, L"$(");
+		size_t startindex = pr->buffer.length;
 		print_embedded_command(pr, wu->wu_cmdsub, indent + 1);
 		wb_wccat(&pr->buffer, L')');
+		if (pr->buffer.contents[startindex] == L'(')
+		    wb_insert(&pr->buffer, startindex, L" ");
 		break;
 	    case WT_ARITH:
 		wb_cat(&pr->buffer, L"$((");
