@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* variable.c: deals with shell variables and parameters */
-/* (C) 2007-2012 magicant */
+/* (C) 2007-2015 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -1538,6 +1538,8 @@ void generate_dirstack_candidates(const le_compopt_T *compopt)
 
 /********** Built-ins **********/
 
+struct reading_option_T;
+
 static void print_variable(
 	const wchar_t *name, const variable_T *var,
 	const wchar_t *argv0, bool readonly, bool export)
@@ -1576,10 +1578,20 @@ static bool set_optind(unsigned long optind, unsigned long optsubind);
 static inline bool set_optarg(const wchar_t *value);
 static bool set_variable_single_char(const wchar_t *varname, wchar_t value)
     __attribute__((nonnull));
-static bool read_with_prompt(xwcsbuf_T *buf, bool noescape)
+static bool read_with_prompt(xwcsbuf_T *buf, xstrbuf_T *split,
+	const struct reading_option_T *ro)
     __attribute__((nonnull));
-static void split_and_assign_array(const wchar_t *name, wchar_t *values,
-	const wchar_t *ifs, bool raw)
+static struct promptset_T promptset_for_read(
+	bool firstline, const struct reading_option_T *ro)
+    __attribute__((nonnull,warn_unused_result));
+static wchar_t *read_one_line_with_prompt(
+	struct promptset_T prompt, bool lineedit)
+    __attribute__((malloc,warn_unused_result));
+static wchar_t *read_one_line(void)
+    __attribute__((malloc,warn_unused_result));
+static bool unescape_line(const wchar_t *line, xwcsbuf_T *buf, xstrbuf_T *split)
+    __attribute__((nonnull));
+static void assign_array(const wchar_t *name, const plist_T *ranges, size_t i)
     __attribute__((nonnull));
 
 /* Options for the "typeset" built-in. */
@@ -2243,10 +2255,8 @@ int unset_builtin(int argc, void **argv)
 	if (function) {
 	    unset_function(name);
 	} else {
-	    if (wcschr(name, L'=')) {
-		xerror(0, Ngt("`%ls' is not a valid variable name"), name);
+	    if (wcschr(name, L'='))
 		continue;
-	    }
 	    unset_variable(name);
 	}
     }
@@ -2534,7 +2544,7 @@ bool check_options(const wchar_t *options)
 		return false;
 	    /* falls thru! */
 	default:
-	    if (posixly_correct && !iswalpha(*options))
+	    if (posixly_correct && !iswalnum(*options))
 		return false;
 	    options++;
 	    if (*options == L':')
@@ -2579,28 +2589,42 @@ const char getopts_syntax[] = Ngt(
 
 /* Options for the "read" built-in. */
 const struct xgetopt_T read_options[] = {
-    { L'A', L"array",    OPTARG_NONE, false, NULL, },
-    { L'r', L"raw-mode", OPTARG_NONE, true,  NULL, },
+    { L'A', L"array",        OPTARG_NONE,     false, NULL, },
+    { L'e', L"line-editing", OPTARG_NONE,     false, NULL, },
+    { L'P', L"ps1",          OPTARG_NONE,     false, NULL, },
+    { L'p', L"prompt",       OPTARG_REQUIRED, false, NULL, },
+    { L'r', L"raw-mode",     OPTARG_NONE,     true,  NULL, },
 #if YASH_ENABLE_HELP
-    { L'-', L"help",     OPTARG_NONE, false, NULL, },
+    { L'-', L"help",         OPTARG_NONE,     false, NULL, },
 #endif
     { L'\0', NULL, 0, false, NULL, },
 };
 
+struct reading_option_T {
+    bool array, lineedit, ps1, raw;
+    const wchar_t *prompt;
+};
+
 /* The "read" built-in, which accepts the following options:
  *  -A: assign values to array
+ *  -e: use line-editing
+ *  -P: use $PS1
+ *  -p: specify prompt
  *  -r: don't treat backslashes specially
  */
 int read_builtin(int argc, void **argv)
 {
-    bool array = false, raw = false;
+    struct reading_option_T ro = { 0 };
 
     const struct xgetopt_T *opt;
     xoptind = 0;
     while ((opt = xgetopt(argv, read_options, 0)) != NULL) {
 	switch (opt->shortopt) {
-	    case L'A':  array = true;  break;
-	    case L'r':  raw   = true;  break;
+	    case L'A':  ro.array    = true;     break;
+	    case L'e':  ro.lineedit = true;     break;
+	    case L'P':  ro.ps1      = true;     break;
+	    case L'p':  ro.prompt   = xoptarg;  break;
+	    case L'r':  ro.raw      = true;     break;
 #if YASH_ENABLE_HELP
 	    case L'-':
 		return print_builtin_help(ARGV(0));
@@ -2610,6 +2634,8 @@ int read_builtin(int argc, void **argv)
 	}
     }
 
+    if (ro.ps1 && ro.prompt != NULL)
+	return mutually_exclusive_option_error(L'P', L'p');
     if (xoptind == argc)
 	return insufficient_operands_error(1);
 
@@ -2621,15 +2647,19 @@ int read_builtin(int argc, void **argv)
 	}
     }
 
-    bool eof;
     xwcsbuf_T buf;
+    xstrbuf_T split;
 
-    /* read input and remove trailing newline */
     wb_init(&buf);
-    if (!read_with_prompt(&buf, raw)) {
+    sb_init(&split);
+    if (!read_with_prompt(&buf, &split, &ro)) {
+	sb_destroy(&split);
 	wb_destroy(&buf);
 	return Exit_FAILURE;
     }
+
+    /* remove trailing newline */
+    bool eof;
     if (buf.length > 0 && buf.contents[buf.length - 1] == L'\n') {
 	wb_truncate(&buf, buf.length - 1);
 	eof = false;
@@ -2638,140 +2668,215 @@ int read_builtin(int argc, void **argv)
 	eof = true;
     }
 
-    const wchar_t *s = buf.contents;
-    const wchar_t *ifs = getvar(L VAR_IFS);
-    plist_T list;
-
     /* split fields */
+    plist_T list;
     pl_init(&list);
-    for (int i = xoptind; i < argc - 1; i++)
-	pl_add(&list, split_next_field(&s, ifs, raw));
-    pl_add(&list, (raw || array) ? xwcsdup(s) : unescape(s));
-    wb_destroy(&buf);
+    {
+	const wchar_t *ifs = getvar(L VAR_IFS);
+	if (ifs == NULL)
+	    ifs = DEFAULT_IFS;
 
-    /* assign variables */
-    assert(xoptind + list.length == (size_t) argc);
-    for (size_t i = 0; i < list.length; i++) {
-	const wchar_t *name = ARGV(xoptind + i);
-	if (i + 1 == list.length)
-	    trim_trailing_spaces(list.contents[i], ifs);
-	if (array && i + 1 == list.length)
-	    split_and_assign_array(name, list.contents[i], ifs, raw);
-	else
-	    set_variable(name, list.contents[i], SCOPE_GLOBAL, shopt_allexport);
+	extract_fields(buf.contents, split.contents, false, ifs, &list);
+	assert(list.length % 2 == 0);
+    }
+
+    /* Add missing empty fields */
+    size_t count = (size_t) argc - xoptind;
+    for (size_t i = list.length / 2; i < count; i++)
+	pl_add(pl_add(&list, buf.contents), buf.contents);
+    assert(list.length % 2 == 0);
+    assert(list.length > 0);
+
+    /* assign variables except last */
+    const wchar_t *name;
+    for (size_t i = 0; i < count - 1; i++) {
+	const wchar_t *start = list.contents[2 * i];
+	const wchar_t *end = list.contents[2 * i + 1];
+	wchar_t *field = xwcsndup(start, end - start);
+	name = ARGV(xoptind + i);
+	set_variable(name, field, SCOPE_GLOBAL, shopt_allexport);
+    }
+
+    /* assign last variable */
+    name = ARGV(xoptind + count - 1);
+    if (ro.array) {
+	assign_array(name, &list, 2 * (count - 1));
+    } else {
+	const wchar_t *start = list.contents[2 * (count - 1)];
+	const wchar_t *end = list.contents[list.length - 1];
+	wchar_t *field = xwcsndup(start, end - start);
+	set_variable(name, field, SCOPE_GLOBAL, shopt_allexport);
     }
 
     pl_destroy(&list);
+    sb_destroy(&split);
+    wb_destroy(&buf);
     return (!eof && yash_error_message_count == 0)
 	    ? Exit_SUCCESS : Exit_FAILURE;
 }
 
-/* Reads input from the standard input and, if `noescape' is false, remove line
- * continuations.
- * The trailing newline and all other backslashes are not removed. */
-bool read_with_prompt(xwcsbuf_T *buf, bool noescape)
+/* Reads one line from the standard input. The result is appended to `buf' and
+ * `split'. `buf' will contain no escapes or other special characters. `split'
+ * is the splittability string for `buf'. The string is splittable at characters
+ * that were not backslash-escaped.
+ * If `ro->raw' is true, exactly one line is read and backslashes are not
+ * treated as escapes. Otherwise, line continuations cause this function to read
+ * more and backslash escapes are recognized.
+ * Returns false on error while reading. */
+bool read_with_prompt(xwcsbuf_T *buf, xstrbuf_T *split,
+	const struct reading_option_T *ro)
 {
-    bool first = true;
-    size_t index;
+    bool firstline = true;
+    bool completed = false;
     bool use_prompt = is_interactive_now && isatty(STDIN_FILENO);
-    struct promptset_T prompt;
 
-read_input:
-    index = buf->length;
-    if (use_prompt) {
-	if (first) {
-	    prompt.main   = xwcsdup(L"");
-	    prompt.right  = xwcsdup(L"");
-	    prompt.styler = xwcsdup(L"");
+    while (!completed) {
+	wchar_t *line;
+	if (use_prompt) {
+	    struct promptset_T prompt = promptset_for_read(firstline, ro);
+	    line = read_one_line_with_prompt(prompt, ro->lineedit);
+	    free_prompt(prompt);
 	} else {
-	    prompt = get_prompt(2);
+	    line = read_one_line();
 	}
+	if (line == NULL)
+	    return false;
 
-#if YASH_ENABLE_LINEEDIT
-	if (shopt_lineedit != SHOPT_NOLINEEDIT) {
-	    wchar_t *line;
-	    inputresult_T result = le_readline(prompt, &line);
-
-	    if (result != INPUT_ERROR) {
-		free_prompt(prompt);
-		switch (result) {
-		    case INPUT_OK:
-			wb_catfree(buf, line);
-			/* falls thru! */
-		    case INPUT_EOF:
-			goto done;
-		    case INPUT_INTERRUPTED:
-			set_interrupted();
-			return false;
-		    case INPUT_ERROR:
-			assert(false);
-		}
-	    }
+	if (ro->raw) {
+	    wb_cat(buf, line);
+	    sb_ccat_repeat(split, true, wcslen(line));
+	    completed = true;
+	} else {
+	    completed = unescape_line(line, buf, split);
 	}
-#endif /* YASH_ENABLE_LINEEDIT */
+	free(line);
 
-	print_prompt(prompt.main);
-	print_prompt(prompt.styler);
+	firstline = false;
     }
-
-    inputresult_T result2 = read_input(buf, stdin_input_file_info, false);
-
-    if (use_prompt) {
-	print_prompt(PROMPT_RESET);
-	free_prompt(prompt);
-    }
-    if (result2 == INPUT_ERROR)
-	return false;
-
-#if YASH_ENABLE_LINEEDIT
-done:
-#endif
-    first = false;
-    if (!noescape) {
-	/* treat escapes */
-	while (index < buf->length) {
-	    if (buf->contents[index] == L'\\') {
-		if (buf->contents[index + 1] == L'\n') {
-		    wb_remove(buf, index, 2);
-		    if (index >= buf->length)
-			goto read_input;
-		    else
-			continue;
-		} else {
-		    index += 2;
-		    continue;
-		}
-	    }
-	    index++;
-	}
-    }
-
     return true;
 }
 
-/* Word-splits `values' and assigns them to the array named `name'.
- * `values' is freed in this function. */
-void split_and_assign_array(const wchar_t *name, wchar_t *values,
-	const wchar_t *ifs, bool raw)
+/* Returns a prompt set for reading in the "read" built-in. */
+struct promptset_T promptset_for_read(
+	bool firstline, const struct reading_option_T *ro)
 {
-    plist_T list;
+    if (!firstline)
+	return get_prompt(2);
+    if (ro->ps1)
+	return get_prompt(1);
 
-    pl_init(&list);
-    if (values[0] != L'\0') {
-	const wchar_t *v = values;
-	while (*v != L'\0')
-	    pl_add(&list, split_next_field(&v, ifs, raw));
+    struct promptset_T ps;
+    ps.main = escape(ro->prompt != NULL ? ro->prompt : L"", L"\\");
+    ps.right = xwcsdup(L"");
+    ps.styler = xwcsdup(L"");
+    return ps;
+}
 
-	if (list.length > 0
-		&& ((wchar_t *) list.contents[list.length - 1])[0] == L'\0') {
-	    free(list.contents[list.length - 1]);
-	    pl_remove(&list, list.length - 1, 1);
+/* Reads one line from the standard input with the specified prompt.
+ * If `lineedit' is true, use line-editing if possible.
+ * The result is returned as a newly-malloced wide string. The result is null
+ * iff an error occurs. */
+wchar_t *read_one_line_with_prompt(struct promptset_T prompt, bool lineedit)
+{
+    wchar_t *line;
+
+    if (lineedit) {
+#if YASH_ENABLE_LINEEDIT
+	if (shopt_lineedit != SHOPT_NOLINEEDIT) {
+	    switch (le_readline(prompt, false, &line)) {
+		case INPUT_OK:
+		    return line;
+		case INPUT_EOF:
+		    return xwcsdup(L"");
+		case INPUT_INTERRUPTED:
+		    set_interrupted();
+		    return NULL;
+		case INPUT_ERROR:
+		    break;
+	    }
 	}
+#endif // YASH_ENABLE_LINEEDIT
     }
 
-    set_array(name, list.length, pl_toary(&list), SCOPE_GLOBAL, false);
+    print_prompt(prompt.main);
+    print_prompt(prompt.styler);
 
-    free(values);
+    line = read_one_line();
+
+    print_prompt(PROMPT_RESET);
+
+    return line;
+}
+
+/* Reads one line from the standard input without printing any prompt or using
+ * line-editing.
+ * The result is returned as a newly-malloced wide string. The result is null
+ * iff an error occurs. */
+wchar_t *read_one_line(void)
+{
+    xwcsbuf_T buf;
+    wb_init(&buf);
+    if (read_input(&buf, stdin_input_file_info, false) != INPUT_ERROR)
+	return wb_towcs(&buf);
+    wb_destroy(&buf);
+    return NULL;
+}
+
+/* Parses a string that may contain backslash escapes.
+ * Unescaped `line' is appended to `buf' with a corresponding splittability
+ * string appended to `split'. Characters are splittable iff not escaped.
+ * The result is false iff `line' ends with a line continuation.
+ * The line continuation is not appended to `buf'. */
+bool unescape_line(const wchar_t *line, xwcsbuf_T *buf, xstrbuf_T *split)
+{
+    for (;;) {
+	bool splitchar;
+
+	switch (*line) {
+	    case L'\0':
+		return true;
+	    case L'\\':
+		line++;
+		switch (*line) {
+		    case L'\0':
+			return true;
+		    case L'\n':
+			return false;
+		}
+		splitchar = false;
+		break;
+	    default:
+		splitchar = true;
+		break;
+	}
+	wb_wccat(buf, *line);
+	sb_ccat(split, splitchar);
+	line++;
+    }
+}
+
+/* Assigns a result of field-splitting contained in `ranges' to an array named
+ * `name'. Fields are assigned starting from index `i' in `ranges'. */
+void assign_array(const wchar_t *name, const plist_T *ranges, size_t i)
+{
+    assert((ranges->length - i) % 2 == 0);
+
+    plist_T fields;
+    pl_init(&fields);
+    while (i < ranges->length) {
+	const wchar_t *start = ranges->contents[i++];
+	const wchar_t *end = ranges->contents[i++];
+	wchar_t *field = xwcsndup(start, end - start);
+	pl_add(&fields, field);
+    }
+
+    if (fields.length == 1 && ((wchar_t *) fields.contents[0])[0] == L'\0') {
+	free(fields.contents[0]);
+	pl_remove(&fields, 0, 1);
+    }
+
+    set_array(name, fields.length, pl_toary(&fields), SCOPE_GLOBAL,
+	    shopt_allexport);
 }
 
 #if YASH_ENABLE_HELP
@@ -2779,7 +2884,7 @@ const char read_help[] = Ngt(
 "read a line from the standard input"
 );
 const char read_syntax[] = Ngt(
-"\tread [-Ar] variable...\n"
+"\tread [-Aer] [-P|-p] variable...\n"
 );
 #endif
 
@@ -2926,7 +3031,7 @@ variable_T *get_dirstack(void)
 void push_dirstack(variable_T *var, wchar_t *value)
 {
     size_t index = var->v_valc++;
-    var->v_vals = xreallocn(var->v_vals, index + 2, sizeof *var->v_vals);
+    var->v_vals = xrealloce(var->v_vals, index, 2, sizeof *var->v_vals);
     var->v_vals[index] = value;
     var->v_vals[index + 1] = NULL;
 }
